# Generated by Django 2.2.14 on 2020-07-17 08:43

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("django_apscheduler", "0003_auto_20200716_1632"),
    ]

    operations = [
        migrations.AlterField(
            model_name="djangojob",
            name="id",
            field=models.CharField(
                help_text="Unique id for this job.",
                max_length=255,
                primary_key=True,
                serialize=False,
            ),
        ),
        migrations.AlterField(
            model_name="djangojobexecution",
            name="status",
            field=models.CharField(
                choices=[
                    ("Started execution", "Started execution"),
                    ("Error!", "Error!"),
                    ("Executed", "Executed"),
                ],
                help_text="The current status of this job execution.",
                max_length=50,
            ),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.12 on 2018-04-12 07:58
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("django_apscheduler", "0001_initial"),
    ]

    operations = [
        migrations.AlterField(
            model_name="djangojob",
            name="next_run_time",
            field=models.DateTimeField(blank=True, db_index=True, null=True),
        ),
    ]



import logging
import pickle
import warnings
from typing import Union, List

from apscheduler import events
from apscheduler.events import JobSubmissionEvent, JobExecutionEvent
from apscheduler.job import Job as AppSchedulerJob
from apscheduler.jobstores.base import BaseJobStore, JobLookupError, ConflictingIdError
from apscheduler.schedulers.base import BaseScheduler

from django import db
from django.db import transaction, IntegrityError

from django_apscheduler.models import DjangoJob, DjangoJobExecution
from django_apscheduler.util import (
    get_apscheduler_datetime,
    get_django_internal_datetime,
)

logger = logging.getLogger(__name__)


class DjangoResultStoreMixin:
    """Mixin class that adds the ability for a JobStore to store job execution results in the Django database"""

    lock = None

    def start(self, scheduler, alias):
        super().start(scheduler, alias)

        # Use the same type of lock as the scheduler to ensure that only one APScheduler event is processed at a time.
        DjangoResultStoreMixin.lock = self._scheduler._create_lock()

        self.register_event_listeners()

    @classmethod
    def handle_submission_event(cls, event: JobSubmissionEvent):
        """
        Create and return new job execution instance in the database when the job is submitted to the scheduler.

        :param event: JobExecutionEvent instance
        :return: DjangoJobExecution ID
        """
        if event.code == events.EVENT_JOB_SUBMITTED:
            # Start logging a new job execution
            job_execution = DjangoJobExecution.atomic_update_or_create(
                cls.lock,
                event.job_id,
                event.scheduled_run_times[0],
                DjangoJobExecution.SENT,
            )
        elif event.code == events.EVENT_JOB_MAX_INSTANCES:
            status = DjangoJobExecution.MAX_INSTANCES

            exception = (
                f"Execution of job '{event.job_id}' skipped: maximum number of running "
                f"instances reached!"
            )

            job_execution = DjangoJobExecution.atomic_update_or_create(
                cls.lock,
                event.job_id,
                event.scheduled_run_times[0],
                status,
                exception=exception,
            )
        else:
            raise NotImplementedError(
                f"Don't know how to handle JobSubmissionEvent '{event.code}'. Expected "
                f"one of '{[events.EVENT_JOB_SUBMITTED, events.EVENT_JOB_MAX_INSTANCES]}'."
            )

        return job_execution.id

    @classmethod
    def handle_execution_event(cls, event: JobExecutionEvent) -> int:
        """
        Store "successful" job execution status in the database.

        :param event: JobExecutionEvent instance
        :return: DjangoJobExecution ID
        """
        if event.code != events.EVENT_JOB_EXECUTED:
            raise NotImplementedError(
                f"Don't know how to handle JobExecutionEvent '{event.code}'. Expected "
                f"'{events.EVENT_JOB_EXECUTED}'."
            )

        job_execution = DjangoJobExecution.atomic_update_or_create(
            cls.lock, event.job_id, event.scheduled_run_time, DjangoJobExecution.SUCCESS
        )
        return job_execution.id

    @classmethod
    def handle_error_event(cls, event: JobExecutionEvent) -> int:
        """
        Store "failed" job execution status in the database.

        :param event: JobExecutionEvent instance
        :return: DjangoJobExecution ID
        """
        if event.code == events.EVENT_JOB_ERROR:

            if event.exception:
                exception = str(event.exception)
                traceback = str(event.traceback)
            else:
                exception = f"Job '{event.job_id}' raised an error!"
                traceback = None

            job_execution = DjangoJobExecution.atomic_update_or_create(
                cls.lock,
                event.job_id,
                event.scheduled_run_time,
                DjangoJobExecution.ERROR,
                exception=exception,
                traceback=traceback,
            )

        elif event.code == events.EVENT_JOB_MISSED:
            # Job execution will not have been logged yet - do so now
            status = DjangoJobExecution.MISSED
            exception = f"Run time of job '{event.job_id}' was missed!"

            job_execution = DjangoJobExecution.atomic_update_or_create(
                cls.lock,
                event.job_id,
                event.scheduled_run_time,
                status,
                exception=exception,
            )

        else:
            raise NotImplementedError(
                f"Don't know how to handle JobExecutionEvent '{event.code}'. Expected "
                f"one of '{[events.EVENT_JOB_ERROR, events.EVENT_JOB_MAX_INSTANCES, events.EVENT_JOB_MISSED]}'."
            )

        return job_execution.id

    def register_event_listeners(self):
        """
        Register various event listeners.

        See: https://github.com/agronholm/apscheduler/blob/master/docs/modules/events.rst for details on which event
        class is used for each event code.

        """
        self._scheduler.add_listener(
            self.handle_submission_event,
            events.EVENT_JOB_SUBMITTED | events.EVENT_JOB_MAX_INSTANCES,
        )

        self._scheduler.add_listener(
            self.handle_execution_event, events.EVENT_JOB_EXECUTED
        )

        self._scheduler.add_listener(
            self.handle_error_event, events.EVENT_JOB_ERROR | events.EVENT_JOB_MISSED,
        )


class DjangoJobStore(DjangoResultStoreMixin, BaseJobStore):
    """
    Stores jobs in a Django database. Based on APScheduler's `MongoDBJobStore`.

    See: https://github.com/agronholm/apscheduler/blob/master/apscheduler/jobstores/mongodb.py

    :param int pickle_protocol: pickle protocol level to use (for serialization), defaults to the
           highest available
    """

    def __init__(self, pickle_protocol: int = pickle.HIGHEST_PROTOCOL):
        super().__init__()
        self.pickle_protocol = pickle_protocol

    def lookup_job(self, job_id: str) -> Union[None, AppSchedulerJob]:
        try:
            job_state = DjangoJob.objects.get(id=job_id).job_state
            return self._reconstitute_job(job_state) if job_state else None

        except DjangoJob.DoesNotExist:
            return None

    def get_due_jobs(self, now) -> List[AppSchedulerJob]:
        dt = get_django_internal_datetime(now)
        return self._get_jobs(next_run_time__lte=dt)

    def get_next_run_time(self):
        try:
            job = DjangoJob.objects.filter(next_run_time__isnull=False).earliest(
                "next_run_time"
            )
            return get_apscheduler_datetime(job.next_run_time, self._scheduler)
        except DjangoJob.DoesNotExist:
            # No active jobs - OK
            return None

    def get_all_jobs(self):
        jobs = self._get_jobs()
        self._fix_paused_jobs_sorting(jobs)

        return jobs

    def add_job(self, job: AppSchedulerJob):
        with transaction.atomic():
            try:
                return DjangoJob.objects.create(
                    id=job.id,
                    next_run_time=get_django_internal_datetime(job.next_run_time),
                    job_state=pickle.dumps(job.__getstate__(), self.pickle_protocol),
                )
            except IntegrityError:
                raise ConflictingIdError(job.id)

    def update_job(self, job: AppSchedulerJob):
        # Acquire lock for update
        with transaction.atomic():
            try:
                db_job = DjangoJob.objects.get(id=job.id)

                db_job.next_run_time = get_django_internal_datetime(job.next_run_time)
                db_job.job_state = pickle.dumps(
                    job.__getstate__(), self.pickle_protocol
                )

                db_job.save()

            except DjangoJob.DoesNotExist:
                raise JobLookupError(job.id)

    def remove_job(self, job_id: str):
        try:
            DjangoJob.objects.get(id=job_id).delete()
        except DjangoJob.DoesNotExist:
            raise JobLookupError(job_id)

    def remove_all_jobs(self):
        # Implicit: will also delete all DjangoJobExecutions due to on_delete=models.CASCADE
        DjangoJob.objects.all().delete()

    def shutdown(self):
        db.connection.close()

    def _reconstitute_job(self, job_state):
        job_state = pickle.loads(job_state)
        job = AppSchedulerJob.__new__(AppSchedulerJob)
        job.__setstate__(job_state)
        job._scheduler = self._scheduler
        job._jobstore_alias = self._alias

        return job

    def _get_jobs(self, **filters):
        jobs = []
        failed_job_ids = set()

        job_states = DjangoJob.objects.filter(**filters).values_list("id", "job_state")
        for job_id, job_state in job_states:
            try:
                jobs.append(self._reconstitute_job(job_state))
            # TODO: Make this except clause more specific
            except Exception:
                self._logger.exception(
                    f"Unable to restore job '{job_id}'. Removing it..."
                )
                failed_job_ids.add(job_id)

        # Remove all the jobs we failed to restore
        if failed_job_ids:
            logger.warning(f"Removing failed jobs: {failed_job_ids}")
            DjangoJob.objects.filter(id__in=failed_job_ids).delete()

        return jobs

    def __repr__(self):
        return f"<{self.__class__.__name__}(pickle_protocol={self.pickle_protocol})>"


def register_events(scheduler, result_storage=None):
    # TODO: Remove this deprecated function in release 0.5
    # DjangoResultStoreMixin now takes care of registering event listeners automatically when the scheduler is started.
    warnings.warn(
        "'register_events' is deprecated since version 0.4.0. Please remove all references from your code.",
        DeprecationWarning,
    )
    pass


def register_job(scheduler: BaseScheduler, *args, **kwargs) -> callable:
    """
    Helper decorator for job registration.

    Automatically fills id parameter to prevent jobs duplication.
    See this comment for explanation: https://github.com/jarekwg/django-apscheduler/pull/9#issuecomment-342074372

    Usage example::

        @register_job(scheduler, "interval", seconds=1)
        def dummy_job():
            print("I'm a job!")

    :param scheduler: Scheduler instance
    :param args, kwargs: Params, will be passed to scheduler.add_job method. See :func:`BaseScheduler.add_job`
    """

    def wrapper_register_job(func):
        kwargs.setdefault("id", f"{func.__module__}.{func.__name__}")
        scheduler.add_job(func, *args, **kwargs)

        return func

    return wrapper_register_job



# Generated by Django 2.2.14 on 2020-07-20 08:48

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("django_apscheduler", "0007_auto_20200717_1404"),
    ]

    operations = [
        migrations.RemoveField(model_name="djangojobexecution", name="started",),
    ]



# Generated by Django 2.2.14 on 2020-07-17 09:07

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("django_apscheduler", "0005_migrate_name_to_id"),
    ]

    operations = [
        migrations.RemoveField(model_name="djangojob", name="name",),
    ]



# Generated by Django 2.2.14 on 2020-07-16 14:32

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ("django_apscheduler", "0002_auto_20180412_0758"),
    ]

    operations = [
        migrations.AlterField(
            model_name="djangojob",
            name="name",
            field=models.CharField(
                help_text="Unique name for this job.", max_length=255, unique=True
            ),
        ),
        migrations.AlterField(
            model_name="djangojob",
            name="next_run_time",
            field=models.DateTimeField(
                blank=True,
                db_index=True,
                help_text="Date and time at which this job is scheduled to be executed next.",
                null=True,
            ),
        ),
        migrations.AlterField(
            model_name="djangojobexecution",
            name="duration",
            field=models.DecimalField(
                decimal_places=2,
                default=None,
                help_text="Total run time of this job (in seconds).",
                max_digits=15,
                null=True,
            ),
        ),
        migrations.AlterField(
            model_name="djangojobexecution",
            name="exception",
            field=models.CharField(
                help_text="Details of exception that occurred during job execution (if any).",
                max_length=1000,
                null=True,
            ),
        ),
        migrations.AlterField(
            model_name="djangojobexecution",
            name="finished",
            field=models.DecimalField(
                decimal_places=2,
                default=None,
                help_text="Timestamp at which this job was finished.",
                max_digits=15,
                null=True,
            ),
        ),
        migrations.AlterField(
            model_name="djangojobexecution",
            name="job",
            field=models.ForeignKey(
                help_text="The job that this execution relates to.",
                on_delete=django.db.models.deletion.CASCADE,
                to="django_apscheduler.DjangoJob",
            ),
        ),
        migrations.AlterField(
            model_name="djangojobexecution",
            name="run_time",
            field=models.DateTimeField(
                db_index=True, help_text="Date and time at which this job was executed."
            ),
        ),
        migrations.AlterField(
            model_name="djangojobexecution",
            name="started",
            field=models.DecimalField(
                decimal_places=2,
                default=None,
                help_text="Timestamp at which this job was started.",
                max_digits=15,
                null=True,
            ),
        ),
        migrations.AlterField(
            model_name="djangojobexecution",
            name="status",
            field=models.CharField(
                choices=[
                    ["Added", "Added"],
                    ["Started execution", "Started execution"],
                    ["Max instances reached!", "Max instances reached!"],
                    ["Missed!", "Missed!"],
                    ["Modified!", "Modified!"],
                    ["Removed!", "Removed!"],
                    ["Error!", "Error!"],
                    ["Executed", "Executed"],
                ],
                help_text="The current status of this job execution.",
                max_length=50,
            ),
        ),
        migrations.AlterField(
            model_name="djangojobexecution",
            name="traceback",
            field=models.TextField(
                help_text="Traceback of exception that occurred during job execution (if any).",
                null=True,
            ),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.1 on 2017-11-05 16:07
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="DjangoJob",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(max_length=255, unique=True)),
                ("next_run_time", models.DateTimeField(db_index=True)),
                ("job_state", models.BinaryField()),
            ],
            options={"ordering": ("next_run_time",),},
        ),
        migrations.CreateModel(
            name="DjangoJobExecution",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "status",
                    models.CharField(
                        choices=[
                            ["Added", "Added"],
                            ["Started execution", "Started execution"],
                            ["Max instances reached!", "Max instances reached!"],
                            ["Missed!", "Missed!"],
                            ["Modified!", "Modified!"],
                            ["Removed!", "Removed!"],
                            ["Error!", "Error!"],
                            ["Executed", "Executed"],
                        ],
                        max_length=50,
                    ),
                ),
                ("run_time", models.DateTimeField(db_index=True)),
                (
                    "duration",
                    models.DecimalField(
                        decimal_places=2, default=None, max_digits=15, null=True
                    ),
                ),
                (
                    "started",
                    models.DecimalField(
                        decimal_places=2, default=None, max_digits=15, null=True
                    ),
                ),
                (
                    "finished",
                    models.DecimalField(
                        decimal_places=2, default=None, max_digits=15, null=True
                    ),
                ),
                ("exception", models.CharField(max_length=1000, null=True)),
                ("traceback", models.TextField(null=True)),
                (
                    "job",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="django_apscheduler.DjangoJob",
                    ),
                ),
            ],
            options={"ordering": ("-run_time",),},
        ),
    ]



from django.db import migrations


def migrate_name_to_id(apps, schema_editor):
    """
    Use DjangoJob.name as the database ID and primary key.
    """
    JobModel = apps.get_model("django_apscheduler", "DjangoJob")
    JobExecutionModel = apps.get_model("django_apscheduler", "DjangoJobExecution")
    migrated_id_mappings = {}
    migrated_job_executions = []

    # Copy 'name' to 'id'.
    for job in JobModel.objects.all():
        migrated_id_mappings[job.id] = job.name
        job.id = job.name
        job.name = f"{job.name}_tmp"
        job.save()

    # Update all job execution references
    for job_execution in JobExecutionModel.objects.filter(
        job_id__in=migrated_id_mappings
    ):
        job_execution.job_id = migrated_id_mappings[job_execution.job_id]
        migrated_job_executions.append(job_execution)

    JobExecutionModel.objects.bulk_update(migrated_job_executions, ["job_id"])

    # Remove old jobs
    JobModel.objects.filter(id__in=migrated_id_mappings).delete()


class Migration(migrations.Migration):

    dependencies = [
        ("django_apscheduler", "0004_auto_20200717_1043"),
    ]

    operations = [migrations.RunPython(migrate_name_to_id)]



from datetime import datetime

from apscheduler.schedulers.base import BaseScheduler
from django.conf import settings
from django.utils import formats
from django.utils import timezone


def get_dt_format() -> str:
    """Return the configured format for displaying datetimes in the Django admin views"""
    return formats.get_format(
        getattr(settings, "APSCHEDULER_DATETIME_FORMAT", "N j, Y, f:s a")
    )


def get_local_dt_format(dt: datetime) -> str:
    """Get the datetime in the localized datetime format"""
    if dt and settings.USE_TZ and timezone.is_aware(dt):
        dt = timezone.localtime(dt)

    return formats.date_format(dt, get_dt_format())


def get_django_internal_datetime(dt: datetime) -> datetime:
    """
    Get the naive or aware version of the datetime based on the configured `USE_TZ` Django setting. This is also the
    format that Django uses to store datetimes internally.
    """
    if dt:
        if settings.USE_TZ and timezone.is_naive(dt):
            return timezone.make_aware(dt)

        elif not settings.USE_TZ and timezone.is_aware(dt):
            return timezone.make_naive(dt)

    return dt


def get_apscheduler_datetime(dt: datetime, scheduler: BaseScheduler) -> datetime:
    """
    Make the datetime timezone aware (if necessary), using the same timezone as is currently configured for the
    scheduler.
    """
    if dt and timezone.is_naive(dt):
        return timezone.make_aware(dt, timezone=scheduler.timezone)

    return dt



# Generated by Django 2.2.14 on 2020-07-17 12:04

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("django_apscheduler", "0006_remove_djangojob_name"),
    ]

    operations = [
        migrations.AlterField(
            model_name="djangojobexecution",
            name="id",
            field=models.BigAutoField(
                help_text="Unique ID for this job execution.",
                primary_key=True,
                serialize=False,
            ),
        ),
    ]



from setuptools import find_packages, setup
from os import path

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

setup(
    name="django-apscheduler",
    version="0.4.2",
    description="APScheduler for Django",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="http://github.com/jarekwg/django-apscheduler",
    author="Jarek Glowacki, Stas Kaledin",
    author_email="jarekwg@gmail.com, staskaledin@gmail.com",
    license="MIT",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Web Environment",
        "Framework :: Django",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Internet :: WWW/HTTP :: Dynamic Content",
    ],
    keywords="django apscheduler django-apscheduler",
    packages=find_packages(exclude=("tests",)),
    install_requires=["django>=2.2", "apscheduler>=3.2",],
    zip_safe=False,
)



from datetime import timedelta, datetime

from django.db import models, transaction
from django.utils import timezone
from django.utils.translation import ugettext_lazy as _

import logging

from django_apscheduler import util
from django_apscheduler.util import get_django_internal_datetime

logger = logging.getLogger(__name__)


class DjangoJob(models.Model):
    id = models.CharField(
        max_length=255, primary_key=True, help_text=_("Unique id for this job.")
    )

    next_run_time = models.DateTimeField(
        db_index=True,
        blank=True,
        null=True,
        help_text=_(
            "Date and time at which this job is scheduled to be executed next."
        ),
    )

    job_state = models.BinaryField()

    def __str__(self):
        status = (
            f"next run at: {util.get_local_dt_format(self.next_run_time)}"
            if self.next_run_time
            else "paused"
        )
        return f"{self.id} ({status})"

    class Meta:
        ordering = ("next_run_time",)


class DjangoJobExecutionManager(models.Manager):
    def delete_old_job_executions(self, max_age: int):
        """
        Delete old job executions from the database.

        :param max_age: The maximum age (in seconds). Executions that are older
        than this will be deleted.
        """
        self.filter(run_time__lte=timezone.now() - timedelta(seconds=max_age)).delete()


class DjangoJobExecution(models.Model):
    SENT = "Started execution"
    SUCCESS = "Executed"
    MISSED = "Missed!"
    MAX_INSTANCES = "Max instances!"
    ERROR = "Error!"

    STATUS_CHOICES = [(x, x) for x in [SENT, ERROR, SUCCESS,]]

    id = models.BigAutoField(
        primary_key=True, help_text=_("Unique ID for this job execution.")
    )

    job = models.ForeignKey(
        DjangoJob,
        on_delete=models.CASCADE,
        help_text=_("The job that this execution relates to."),
    )

    status = models.CharField(
        max_length=50,
        # TODO: Replace this with enumeration types when we drop support for Django 2.2
        # See: https://docs.djangoproject.com/en/dev/ref/models/fields/#field-choices-enum-types
        choices=STATUS_CHOICES,
        help_text=_("The current status of this job execution."),
    )

    run_time = models.DateTimeField(
        db_index=True, help_text=_("Date and time at which this job was executed."),
    )

    # We store this value in the DB even though it can be calculated as `finished - run_time`. This allows quick
    # calculation of average durations directly in the database later on.
    duration = models.DecimalField(
        max_digits=15,
        decimal_places=2,
        default=None,
        null=True,
        help_text=_("Total run time of this job (in seconds)."),
    )

    finished = models.DecimalField(
        max_digits=15,
        decimal_places=2,
        default=None,
        null=True,
        help_text=_("Timestamp at which this job was finished."),
    )

    exception = models.CharField(
        max_length=1000,
        null=True,
        help_text=_(
            "Details of exception that occurred during job execution (if any)."
        ),
    )

    traceback = models.TextField(
        null=True,
        help_text=_(
            "Traceback of exception that occurred during job execution (if any)."
        ),
    )

    objects = DjangoJobExecutionManager()

    @classmethod
    def atomic_update_or_create(
        cls,
        lock,
        job_id: str,
        run_time: datetime,
        status: str,
        exception: str = None,
        traceback: str = None,
    ) -> "DjangoJobExecution":
        """
        Uses an APScheduler lock to ensures that only one database entry can be created / updated at a time.

        This keeps django_apscheduler in sync with APScheduler and maintains a 1:1 mapping between APScheduler events
        that are triggered and the corresponding DjangoJobExecution model instances that are persisted to the database.
        :param lock: The lock to use when updating the database - probably obtained by calling _scheduler._create_lock()
        :param job_id: The ID to the APScheduler job that this job execution is for.
        :param run_time: The scheduler runtime for this job execution.
        :param status: The new status for ths job execution.
        :param exception: Details of any exceptions that need to be logged.
        :param traceback: Traceback of any exceptions that occurred while executing the job.
        :return: The ID of the newly created or updated DjangoJobExecution.
        """

        # Ensure that only one update / created can be processed at a time, staying in sync with corresponding
        # scheduler.
        with lock:
            # Convert all datetimes to internal Django format before doing calculations and persisting in the database.
            run_time = get_django_internal_datetime(run_time)

            finished = get_django_internal_datetime(timezone.now())
            duration = (finished - run_time).total_seconds()
            finished = finished.timestamp()

            try:
                with transaction.atomic():
                    job_execution = DjangoJobExecution.objects.select_for_update().get(
                        job_id=job_id, run_time=run_time
                    )

                    if status == DjangoJobExecution.SENT:
                        # Ignore 'submission' events for existing job executions. APScheduler does not appear to
                        # guarantee the order in which events are received, and it is not unusual to receive an
                        # `executed` before the corresponding `submitted` event. We just discard `submitted` events
                        # that are received after the job has already been executed.
                        #
                        # If there are any more instances like this then we probably want to implement a full blown
                        # state machine using something like `pytransitions`
                        # See https://github.com/pytransitions/transitions
                        return job_execution

                    job_execution.finished = finished
                    job_execution.duration = duration
                    job_execution.status = status

                    if exception:
                        job_execution.exception = exception

                    if traceback:
                        job_execution.traceback = traceback

                    job_execution.save()

            except DjangoJobExecution.DoesNotExist:
                # Execution was not created by a 'submit' previously - do so now
                if status == DjangoJobExecution.SENT:
                    # Don't log durations until after job has been submitted for execution
                    finished = None
                    duration = None

                job_execution = DjangoJobExecution.objects.create(
                    job_id=job_id,
                    run_time=run_time,
                    status=status,
                    duration=duration,
                    finished=finished,
                    exception=exception,
                    traceback=traceback,
                )

        return job_execution

    def __str__(self):
        return f"{self.id}: job '{self.job_id}' ({self.status})"

    class Meta:
        ordering = ("-run_time",)



from django.contrib import admin
from django.db.models import Avg
from django.utils.safestring import mark_safe

from django_apscheduler.models import DjangoJob, DjangoJobExecution
from django_apscheduler import util


@admin.register(DjangoJob)
class DjangoJobAdmin(admin.ModelAdmin):
    search_fields = ["id"]
    list_display = ["id", "local_run_time", "average_duration"]

    def get_queryset(self, request):
        qs = super().get_queryset(request)

        self.avg_duration_qs = (
            DjangoJobExecution.objects.filter(
                job_id__in=qs.values_list("id", flat=True)
            )
            .order_by("job_id")
            .values_list("job")
            .annotate(avg_duration=Avg("duration"))
        )

        return qs

    def local_run_time(self, obj):
        if obj.next_run_time:
            return util.get_local_dt_format(obj.next_run_time)

        return "(paused)"

    def average_duration(self, obj):
        try:
            return self.avg_duration_qs.get(job_id=obj.id)[1]
        except DjangoJobExecution.DoesNotExist:
            return "None"

    average_duration.short_description = "Average Duration (sec)"


@admin.register(DjangoJobExecution)
class DjangoJobExecutionAdmin(admin.ModelAdmin):
    status_color_mapping = {
        DjangoJobExecution.SUCCESS: "green",
        DjangoJobExecution.SENT: "blue",
        DjangoJobExecution.MAX_INSTANCES: "orange",
        DjangoJobExecution.MISSED: "orange",
        DjangoJobExecution.ERROR: "red",
    }

    list_display = ["id", "job", "html_status", "local_run_time", "duration_text"]
    list_filter = ["job__id", "run_time", "status"]

    def html_status(self, obj):
        return mark_safe(
            f'<p style="color: {self.status_color_mapping[obj.status]}">{obj.status}</p>'
        )

    def local_run_time(self, obj):
        return util.get_local_dt_format(obj.run_time)

    def duration_text(self, obj):
        return obj.duration or "N/A"

    html_status.short_description = "Status"
    duration_text.short_description = "Duration (sec)"
